import os
import re

def is_windows_path(path: str) -> bool:
    """Detects if a given path is a Windows path."""
    return bool(re.match(r"^[a-zA-Z]:\\", path))  # Matches "C:\", "D:\", etc.

def convert_to_wsl_path(path: str) -> str:
    """Converts a Windows path to a WSL path if running inside WSL."""
    if "WSL_DISTRO_NAME" in os.environ and is_windows_path(path):
        drive, rest = path[0].lower(), path[2:].replace("\\", "/")
        return f"/mnt/{drive}{rest}"
    return path  # Return unchanged if not in WSL or not a Windows path

def convert_to_windows_path(wsl_path: str) -> str:
    """Converts a WSL path to a Windows path if running inside WSL."""
    if wsl_path.startswith("/mnt/") and len(wsl_path) > 6:
        drive, rest = wsl_path[5], wsl_path[7:].replace("/", "\\")
        return f"{drive.upper()}:\\{rest}"
    return wsl_path  # Return unchanged if not a WSL path



from typing import Any
import parse #type:ignore

class StringParser:
    """ Class for converting string -> dict and vice versa given a list of filename formats """

    def __init__(self, formats: list[str]|str|None, use_tags:bool=True):
        if formats == None:
            raise Exception('No format(s) given to the Parser')
        self.formats = self._expand_formats(formats)
        self.use_tags = use_tags
        self.tags_sep = ' #'
    
    def parse(self, string: str) -> dict[str, Any] | None:
        for f in self.formats:
            string_cpy = string
            extracted_tags = []
            if self.use_tags:
                string_cpy, extracted_tags = self._extract_tags_from_string(string_cpy)
            parsed_data = parse.parse(f, string_cpy) #type:ignore
            if parsed_data != None:
                data = parsed_data.named #type:ignore
                if self.use_tags:
                    data['tags'] = extracted_tags
                return data #type:ignore
        return None
    
    def format(self, data: dict[str, Any]) -> str | None:
        data = self._prune_data(data)
        tags, data = self._separate_tags(data)
        for f in self.formats:
            f = self._remove_unsupported_format_codes(f)
            try:
                str = f.format(**data)
                if self.use_tags:
                    for tag in tags:
                        str += self.tags_sep + tag
                return str
            except KeyError:
                pass
        return None
    
    def _extract_tags_from_string(self, string: str):
        """ Extracts tags defined as `#TagName` from a string """
        tags: list[str] = []
        parts = string.split(self.tags_sep)
        for i in reversed(range(1, len(parts))):
            part = parts[i]
            if not ' ' in part:
                tags.append(part)
        return parts[0], tags
    
    @staticmethod
    def _separate_tags(data: dict[str, Any]):
        tags = []
        if 'tags' in data:
            tags = [ t.replace(' ', '-') for t in data['tags'] ]
            del data['tags']
        return tags, data
    
    @staticmethod
    def _remove_unsupported_format_codes(fmt: str):
        for code in [':S', ':D']:
            fmt = fmt.replace(code, '')
        return fmt
    
    def _expand_formats(self, formats: list[str] | str) -> list[str]:
        opt_sig = ';opt'
        if not isinstance(formats, list):
            formats = [formats]
        new_formats: list[str] = []
        for fmt_base in formats:
            format_parts = fmt_base.split()
            optional_count = len([c for c in format_parts if c.endswith(opt_sig)])
            for n in range(2**optional_count):
                parts: list[str] = []
                i = 0
                for part in format_parts:
                    if not part.endswith(opt_sig):
                        parts.append(part)
                    else:
                        mask = 2 ** i
                        if n & mask:
                            parts.append(part.replace(opt_sig, ''))
                        i += 1
                fmt = ' '.join(parts)
                new_formats.append(fmt)
        new_formats.sort(
            reverse=True, 
            key=lambda fmt: ( len(self._get_parse_in_fmt(fmt)), len(self._get_non_param_chars(fmt)) )
        )
        return new_formats
    
    @staticmethod
    def _get_parse_in_fmt(fmt: str) -> list[str]:
        return fmt.split('}')
    
    @staticmethod
    def _get_non_param_chars(fmt: str) -> str:
        parts = fmt.split('{')
        parts = [ p.split('}')[-1] for p in parts ]
        string = ''.join(parts).replace(' ', '')
        return string
    
    @staticmethod
    def _is_date(str: str):
        for c in '.-_':
            str = str.replace(c, '')
        return str.isnumeric()
    
    @staticmethod
    def _prune_data(data: dict[str, Any]):
        remove_keys = [ k for k, v in data.items() if v == None ]
        for k in remove_keys:
            del data[k]
        return data

    @staticmethod
    def _to_cc(string: str):
        if ' ' not in string:
            return string
        parts = [ p for p in string.lower().split(" ") if p != '' ]
        for i in range(len(parts)):
            part = parts[i]
            parts[i] = part[:1].upper() + part[1:]
        return ''.join(parts)
    
    @staticmethod
    def _from_cc(string: str):
        chars: list[str] = []
        for c in list(string):
            if c.isupper():
                chars.append(' ')
            chars.append(c)
        return ''.join(chars)




# Ideas for v2:
# - add temp .~lock file when open (thing about this carefully, eg what happens when program crashes, how to use class instance id or something)
# - handler can be treated like a dict (handler[key] = value etc)
#   * handler[key] = value
#   * del handler[key]
#   * key in handler
#   * len(handler)
# - simplify methods    .getValue() -> .get()
# - add removeKey() (AND del handler[key])
# - add iteration of items without needing .items()
from typing import Any
import json
import os
import shutil
from datetime import datetime

class JsonHandler:
    """ A class for abtrsacting the reading and writing process of json files/data """
    
    def __init__(self, filepath: str, readonly:bool=False, prettify:bool=False):
        self.readonly = readonly
        self.prettify = prettify
        self.curdir = os.path.abspath(os.curdir)
        if ":" in filepath:
            self.filepath = filepath
        else:
            self.filepath = os.path.join(self.curdir, filepath)
        filedir = os.path.dirname(self.filepath)
        self.backupdir = os.path.join(filedir, "BAK")
        self.jsonObject = self.load()
    
    def addItem(self, key: Any, value: Any, nosave:bool=False):
        if self.hasKey(key):
            return False
        self.jsonObject[key] = value
        if not nosave:
            self.save()
        return True
    
    def setValue(self, key: Any, value: Any, nosave:bool=False):
        self.jsonObject[key] = value
        if not nosave:
            self.save()
        return True
    
    def appendValue(self, key: Any, value: Any, nosave:bool=False):
        if key in self.jsonObject:
            if not isinstance(self.jsonObject[key], list):
                return False
        else:
            self.jsonObject[key] = []
        self.jsonObject[key].append(value)
        if not nosave:
            self.save()
        return True

    def getKeys(self):
        return list(self.jsonObject.keys())
    
    def hasKey(self, key: Any):
        return key in self.jsonObject
    
    def getValues(self):
        return list(self.jsonObject.values())

    def getValue(self, key: Any, noValueRet:Any=None):
        if key not in self.jsonObject:
            return noValueRet
        return self.jsonObject[key]

    def getItems(self) -> list[tuple[Any, Any]]:
        return list(self.jsonObject.items())
    
    def load(self) -> dict[Any, Any]:
        #print("Loading json ...")
        try:
            with open(self.filepath, 'r') as openfile:
                jsonObject = json.load(openfile)
            return jsonObject
        except:
            return {}
        
    def save(self):
        if self.readonly:
            print("WARNING: Unable to save, handler in READONLY mode")
            return
        temppath = self.filepath + '.tmp'
        with open(temppath, "w") as outfile:
            if self.prettify:
                outfile.write(json.dumps(self.jsonObject, indent=4))
            else:
                outfile.write(json.dumps(self.jsonObject))
        os.replace(temppath, self.filepath)
        
    def backup(self):
        if not os.path.exists(self.backupdir):
            os.mkdir(self.backupdir)
            print("Made dir")
        savename = os.path.basename(self.filepath) + "_BAK_" + str(datetime.now().strftime("%y-%m-%d_%H-%M-%S")) + ".json"
        savepath = os.path.join(self.backupdir, savename)
        print("Saving backup to: ", savepath)
        shutil.copyfile(self.filepath, savepath)




from typing import Any
import os
import json
import datetime


class BookmarksGetter:
    '''
    Class for getting bookmarks from your browser into your python code.
    
    Supported browsers:
    - Brave
    - Chrome
    '''
    DEFUALT_BOOKMARKS_PATHS = [
        r'%localappdata%\Google\Chrome\User Data\Default\Bookmarks',
        r'%localappdata%\BraveSoftware\Brave-Browser\User Data\Default\Bookmarks',
    ]

    @classmethod
    def get_bookmarks(cls, browser: str, foldername:str|None=None, domain:str|list[str]|None=None, sortby:str|None=None, reverse:bool=False) -> list[dict[str, Any]] | None:
        """ Get bookmarks from broswer as a list of python dicts """
        if domain and not isinstance(domain, list):
            domain = [domain]
        bookmarks_path = next( (pth for pth in cls.DEFUALT_BOOKMARKS_PATHS if (browser.lower() in pth.lower() )) , None)
        if bookmarks_path == None:
            print('Cannot find bookmarks for browser named "{}"'.format(browser))
            return None
        bookmarks_path = bookmarks_path.replace(r'%localappdata%', os.getenv('LOCALAPPDATA')) #type:ignore
        with open(bookmarks_path, 'r') as f:
            bookmarks_json = json.load(f)
        if browser.lower() not in 'chrome chromium bravesoftware edge':
            print('ERROR: "{}" is in an unknown browser family'.format(browser))
            return None
        bookmarks_objects = cls._get_bookmarks_Chrome(bookmarks_json, foldername=foldername)
        bookmarks_objects = [ b for b in bookmarks_objects if ( foldername==None    or foldername.lower() in b.get('location','').lower().split('/')) ]
        bookmarks_objects = [ b for b in bookmarks_objects if ( domain==None        or 0 != len([dom for dom in domain if dom.lower() in b.get('url','').lower()]) ) ]
        for bm in bookmarks_objects: bm['location_relative'] = cls._get_relative_bookmark_location(bm['location'], foldername)
        if sortby and bookmarks_objects != []:
            if sortby not in bookmarks_objects[0].keys():
                raise Exception('ERROR in BookmarksGetter: No such key "{}"'.format(sortby))
            else:
                bookmarks_objects.sort( key=lambda obj: obj.get(sortby, ''), reverse=reverse )
        else:
            bookmarks_objects.sort( key=lambda obj: obj.get('date_added', ''), reverse=reverse )
        return bookmarks_objects

    # TODO: figure out foldername
    @classmethod
    def _get_bookmarks_Chrome(cls, bookmarks_json: dict[Any, Any], foldername:str|None=None):
        base_objects = bookmarks_json['roots']['bookmark_bar'].get('children')
        return cls._get_bookmarks_as_list_Chrome(base_objects)
    
    @classmethod
    def _get_bookmarks_as_list_Chrome(cls, array:list[dict[str, Any]], location:str|None=None):
        bookmarks:list[dict[str, Any]] = []
        for obj in array:
            if obj.get('type') == 'url':
                obj['location'] = location if location else ''
                obj['date_added_fmt'] = cls._windows_epoch_readable(obj['date_added'])
                obj['date_last_used_fmt'] = cls._windows_epoch_readable(obj['date_last_used'])
                if obj.get('date_modified'):
                    obj['date_modified_fmt'] = cls._windows_epoch_readable(obj['date_modified'])
                bookmarks.append(obj)
            elif obj.get('type') == 'folder':
                name = obj.get('name')
                children = obj.get('children', [])
                new_location = f'{location}/{name}' if location else name
                bookmarks.extend(cls._get_bookmarks_as_list_Chrome(children, new_location))
        return bookmarks


    ### HELPER METHODS
    @staticmethod
    def _windows_epoch_readable(us: str) -> str:
        windows_epoch_start = datetime.datetime(1601, 1, 1)
        return str(windows_epoch_start + datetime.timedelta(microseconds=int(us)))
    
    @staticmethod
    def _get_relative_bookmark_location(location:str, foldername:str|None):
        if foldername == None:
            return location
        if location == foldername:
            return ''
        return location.replace(f'{foldername}/', '')



""" A collection of small miscellaneous libraries by MattThePerson (https://github.com/MattThePerson/handymatt) """
from .string_parser import StringParser
from .json_handler import JsonHandler
from .bookmarks_getter import BookmarksGetter
import handymatt.wsl_paths as wsl_paths

__all__ = ["StringParser", "JsonHandler", "BookmarksGetter", "wsl_paths"]
